// Copyright 2007 Hiroyuki Ogasawara
// vim:ts=4 sw=4:
/*!
	@file
	@date	2007/10/24
	@author	Hiroyuki Ogasawara
	@brief

	`̃sNZ\[g

	@sa http://dench.flatlib.jp/
	@sa http://wlog.flatlib.jp/
*/


#include	"whsample.h"


#define		_APP_TITLE_		TEXT("Wheel Handle ss10   http://dench.flatlib.jp")



/*!
	@brief
		Tvt[[N
*/
class Sample : public whSample {
public:
	enum {
		WINDOW_SIZE_WIDTH	= 640,
		WINDOW_SIZE_HEIGHT	= 480,
		MAX_RENDERBUFFER	= 4,
		MAX_TEAPOTS			= 8,
		MIN_TEAPOTS			= 1,
	};
private:
	D3DXMATRIX					ViewMatrix;
	D3DXMATRIX					Projection;
	ID3D10Effect*				iEffectModel;
	ID3D10Buffer*				iVertexBuffer;
	ID3D10Buffer*				iIndexBuffer;
	ID3D10InputLayout*			iInputLayout;
	ID3D10BlendState*			iBlendState;
	ID3D10BlendState*			iBlendDisable;
	ID3D10RenderTargetView*		iRenderBuffer[MAX_RENDERBUFFER];
	ID3D10ShaderResourceView*	iShaderResourceView[MAX_RENDERBUFFER];
	float			Frame;
	float			RotateAngle;
	unsigned int	Sorting;
	unsigned int	Animation;
	unsigned int	LayerMode;
	unsigned int	Teapots;
public:
	void	InitShader();
	void	ReleaseShader();
	void	Render();

	Sample() :
		iEffectModel( NULL ),
		iVertexBuffer( NULL ),
		iIndexBuffer( NULL ),
		iInputLayout( NULL ),
		iBlendState( NULL ),
		iBlendDisable( NULL ),
		Frame( 0.0f ),
		RotateAngle( 0.0f ),
		LayerMode( 6 ),
		Teapots( 4 ),
		Sorting( TRUE ),
		Animation( TRUE )
	{
		for( int i= 0 ; i< MAX_RENDERBUFFER ; i++ ){
			iRenderBuffer[i]= NULL;
			iShaderResourceView[i]= NULL;
		}
		InitDevice( TRUE, WINDOW_SIZE_WIDTH, WINDOW_SIZE_HEIGHT, _APP_TITLE_ );
		InitShader();
	}
	~Sample()
	{
		ReleaseShader();
		ReleaseDevice();
	}
	void	wmChar( WPARAM wparam, LPARAM lparam )
	{
		switch( wparam ){
		case '6':
		case '2':
			LayerMode= 6;
			Teapots= 4;
			break;
		case '3':
		case '1':
			LayerMode= 3;
			Teapots= 2;
			break;
		case 'u':
		case 'U':
			if( ++Teapots > MAX_TEAPOTS ){
				Teapots= MAX_TEAPOTS;
			}
			break;
		case 'd':
		case 'D':
			if( --Teapots < MIN_TEAPOTS ){
				Teapots= MIN_TEAPOTS;
			}
			break;
		case 'S':
		case 's':
		case '\t':
			Sorting^= 1;
			break;
		case '\r':
		case '\n':
		case ' ':
			Animation^= 1;
			break;
		}
	}
};


/*
	ff[^
*/

struct VertexType	{
	float	x, y, z;
	float	nx, ny, nz;
	float	tu, tv;
};
typedef	unsigned short	IndexType;

#include	"teapot.inc"


/*!
	@brief
		
*/
void
Sample::InitShader()
{

	// Shader ̓ǂݍ
	ID3D10Blob*	iblob= NULL;
	if( FAILED( D3DX10CreateEffectFromFile(
				TEXT("model.fxo"),
				NULL, NULL, "fx_4_0", 0, 0, iDevice,
				NULL, NULL, &iEffectModel, &iblob, NULL ) ) ){
		OutputDebugStringA( "\n" );
		OutputDebugStringA( reinterpret_cast<char*>(
									iblob->GetBufferPointer() ) );
		OutputDebugStringA( "\n" );
		ZRELEASE( iblob );
		return;
	}



	// InputLayout
	static const D3D10_INPUT_ELEMENT_DESC	_inputDesc[]= {
		{	"POSITION",	0,	DXGI_FORMAT_R32G32B32_FLOAT,	0,
					0,					D3D10_INPUT_PER_VERTEX_DATA, 0,	},
		{	"NORMAL",	0,	DXGI_FORMAT_R32G32B32_FLOAT,	0,
					sizeof(float)*3,	D3D10_INPUT_PER_VERTEX_DATA, 0,	},
		{	"TEXCOORD",	0,	DXGI_FORMAT_R32G32_FLOAT,		0,
					sizeof(float)*6,	D3D10_INPUT_PER_VERTEX_DATA, 0,	},
	};
	D3D10_PASS_DESC	pdesc;
	iEffectModel->GetTechniqueByIndex(0)->GetPassByIndex(0)->GetDesc( &pdesc );
	iDevice->CreateInputLayout( _inputDesc, 3,
			pdesc.pIAInputSignature, pdesc.IAInputSignatureSize,
			&iInputLayout );



	// VertexBuffer
	D3D10_BUFFER_DESC	bdesc= {
		VertexCount * sizeof(VertexType),
		D3D10_USAGE_IMMUTABLE,
		D3D10_BIND_VERTEX_BUFFER,
		0,
		0,
	};
	D3D10_SUBRESOURCE_DATA	initdata= { NULL, 0, 0, };
	initdata.pSysMem= _Vertex;
	iDevice->CreateBuffer( &bdesc, &initdata, &iVertexBuffer );


	// IndexBuffer
	bdesc.ByteWidth= IndexCount * sizeof(IndexType);
	bdesc.BindFlags= D3D10_BIND_INDEX_BUFFER;
	initdata.pSysMem= _Index;
	iDevice->CreateBuffer( &bdesc, &initdata, &iIndexBuffer );


	// Camera
	D3DXMATRIX	projection;
	D3DXMatrixPerspectiveFovLH( &Projection,
				3.141592f/3.0f,
				WINDOW_SIZE_WIDTH/(float)WINDOW_SIZE_HEIGHT,
				1.0f, 1000.0f );

	D3DXMATRIX	viewmatrix;
	D3DXVECTOR3	eye( 0.0f, 16.0f, -30.0f );
	D3DXVECTOR3	at( 0.0f, 0.0f, 0.0f );
	D3DXVECTOR3	up( 0.0f, 1.0f, 0.0f );
	D3DXMatrixLookAtLH( &ViewMatrix, &eye, &at, &up );

	D3DXMATRIX	viewprojection;
	D3DXMatrixMultiply( &viewprojection, &ViewMatrix, &Projection );
	iEffectModel->GetVariableByName( "ViewProjection" )->AsMatrix()
					->SetMatrix( &viewprojection._11 );
	iEffectModel->GetVariableByName( "CameraPosition" )->AsVector()
					->SetFloatVector( &eye.x );


	// BlendState
	// 
	static const D3D10_BLEND_DESC	blenddesc= {
		FALSE,
		{	TRUE,	TRUE,	TRUE,	TRUE,	FALSE,	FALSE,	FALSE,	FALSE,	},
		D3D10_BLEND_DEST_ALPHA,
		D3D10_BLEND_ONE,
		D3D10_BLEND_OP_ADD,
		D3D10_BLEND_ZERO,
		D3D10_BLEND_SRC_ALPHA, //D3D10_BLEND_BLEND_FACTOR,
		D3D10_BLEND_OP_REV_SUBTRACT,
		{
			D3D10_COLOR_WRITE_ENABLE_ALL,
			D3D10_COLOR_WRITE_ENABLE_ALL,
			D3D10_COLOR_WRITE_ENABLE_ALL,
			D3D10_COLOR_WRITE_ENABLE_ALL,
			D3D10_COLOR_WRITE_ENABLE_ALL,
			D3D10_COLOR_WRITE_ENABLE_ALL,
			D3D10_COLOR_WRITE_ENABLE_ALL,
			D3D10_COLOR_WRITE_ENABLE_ALL,
		},
	};
	iDevice->CreateBlendState( &blenddesc, &iBlendState );


	// s
	static const D3D10_BLEND_DESC	blenddesc2= {
		FALSE,
		{	FALSE,	FALSE,	FALSE,	FALSE,	FALSE,	FALSE,	FALSE,	FALSE,	},
		D3D10_BLEND_ONE,
		D3D10_BLEND_ONE,
		D3D10_BLEND_OP_ADD,
		D3D10_BLEND_ONE,
		D3D10_BLEND_ONE,
		D3D10_BLEND_OP_ADD,
		{
			D3D10_COLOR_WRITE_ENABLE_ALL,
			D3D10_COLOR_WRITE_ENABLE_ALL,
			D3D10_COLOR_WRITE_ENABLE_ALL,
			D3D10_COLOR_WRITE_ENABLE_ALL,
			D3D10_COLOR_WRITE_ENABLE_ALL,
			D3D10_COLOR_WRITE_ENABLE_ALL,
			D3D10_COLOR_WRITE_ENABLE_ALL,
			D3D10_COLOR_WRITE_ENABLE_ALL,
		},
	};
	iDevice->CreateBlendState( &blenddesc2, &iBlendDisable );


	// target
	for( int i= 0 ; i< MAX_RENDERBUFFER ; i++ ){
		ID3D10Texture2D*	riTexture2D= NULL;
		D3D10_TEXTURE2D_DESC	tex2ddesc= {
			WINDOW_SIZE_WIDTH,
			WINDOW_SIZE_HEIGHT,
			1,	// mip
			1,	// array
			DXGI_FORMAT_R32G32B32A32_FLOAT,
			{ 1, 0,	},
			D3D10_USAGE_DEFAULT,
			D3D10_BIND_SHADER_RESOURCE | D3D10_BIND_RENDER_TARGET,
			0,
			0,
		};
		if( FAILED( iDevice->CreateTexture2D(
					&tex2ddesc,
					NULL,
					&riTexture2D
				) ) ){
			OutputDebugStringA( "CreateTexture2D\n" );
		}

		iShaderResourceView[i]= NULL;
		D3D10_SHADER_RESOURCE_VIEW_DESC	srvdesc;
		srvdesc.Format= DXGI_FORMAT_R32G32B32A32_FLOAT;
		srvdesc.ViewDimension= D3D10_SRV_DIMENSION_TEXTURE2D;
		srvdesc.Texture2D.MostDetailedMip= 0;
		srvdesc.Texture2D.MipLevels= 1;
		if( FAILED( iDevice->CreateShaderResourceView(
					riTexture2D,
					&srvdesc,
					&iShaderResourceView[i]
					) ) ){
			OutputDebugStringA( "CreateShaderResourceView\n" );
		}

		iRenderBuffer[i]= NULL;
		D3D10_RENDER_TARGET_VIEW_DESC	rtvdesc;
		rtvdesc.Format= DXGI_FORMAT_R32G32B32A32_FLOAT;
		rtvdesc.ViewDimension= D3D10_RTV_DIMENSION_TEXTURE2D;
		rtvdesc.Texture2D.MipSlice= 0;
		if( FAILED( iDevice->CreateRenderTargetView(
					riTexture2D,
					&rtvdesc,
					&iRenderBuffer[i]
					) ) ){
			OutputDebugStringA( "CreateRenderTargetView\n" );
		}

		ZRELEASE( riTexture2D );
	}

}


/*!
	@brief
		I
*/
void
Sample::ReleaseShader()
{
	if( iDevice ){
		iDevice->ClearState();
	}

	for( int i= 0 ; i< MAX_RENDERBUFFER ; i++ ){
		ZRELEASE( iShaderResourceView[i] );
		ZRELEASE( iRenderBuffer[i] );
	}
	ZRELEASE( iBlendState );
	ZRELEASE( iBlendDisable );
	ZRELEASE( iIndexBuffer );
	ZRELEASE( iVertexBuffer );
	ZRELEASE( iInputLayout );
	ZRELEASE( iEffectModel );
}



/*!
	@brief
		Rendering
*/
void
Sample::Render()
{
	const char*	_TechniqueMain= "Main6";
	const char*	_TechniqueSort= "Sort6";

	if( LayerMode == 3 ){
		_TechniqueMain= "Main3";
		_TechniqueSort= "Sort3";
	}

	// Animation
	if( Animation ){
		RotateAngle+= 0.02f;
		Frame+= 1.0f;
	}


	// ʃNA
	static const float	color[4]= {
		0.0f, 0.0f, 1.0f, 1.0f,	// dv
	};
	for( int i= 0 ; i< MAX_RENDERBUFFER ; i++ ){
		iDevice->ClearRenderTargetView( iRenderBuffer[i], color );
	}
	iDevice->ClearDepthStencilView( iDepthStencilView,
				D3D10_CLEAR_DEPTH|D3D10_CLEAR_STENCIL, 1.0f, 0 );


	// MRTݒ
	iDevice->OMSetRenderTargets(
				MAX_RENDERBUFFER, iRenderBuffer, iDepthStencilView );


	// XV
	D3DXMATRIX	worldmatrix;
	D3DXMatrixRotationY( &worldmatrix, RotateAngle );
	iEffectModel->GetVariableByName( "World" )->AsMatrix()
								->SetMatrix( &worldmatrix._11 );
	iEffectModel->GetVariableByName( "cFrame" )->AsScalar()->SetFloat( Frame );
	iEffectModel->GetVariableByName( "Sorting" )->AsScalar()->SetInt( Sorting );


	// Blend ݒ
	static const float	bfactor[4]= {
		256.0f, 256.0f, 256.0f, 256.0f,
	};
	iDevice->OMSetBlendState( iBlendState, bfactor, 0xffffffff );


	// f̕`
	iDevice->IASetInputLayout( iInputLayout );
	UINT	vsize= sizeof(VertexType);
	UINT	voffset= 0;
	iDevice->IASetVertexBuffers( 0, 1, &iVertexBuffer, &vsize, &voffset );
	iDevice->IASetIndexBuffer( iIndexBuffer, DXGI_FORMAT_R16_UINT, 0 );
	iEffectModel->GetTechniqueByName( _TechniqueMain )->
									GetPassByIndex( 0 )->Apply( 0 );

	iDevice->IASetPrimitiveTopology( D3D10_PRIMITIVE_TOPOLOGY_TRIANGLELIST );
	iDevice->DrawIndexedInstanced( IndexCount, Teapots, 0, 0, 0 );


	// Blend ݒ߂
	iDevice->OMSetBlendState( iBlendDisable, bfactor, 0xffffffff );


	// back buffer ֓]
	iDevice->IASetPrimitiveTopology( D3D10_PRIMITIVE_TOPOLOGY_TRIANGLESTRIP );
	ID3D10Buffer*	_nullBuf= NULL;
	iDevice->IASetVertexBuffers( 0, 1, &_nullBuf, &vsize, &voffset );
	iDevice->IASetInputLayout( NULL );

	iDevice->OMSetRenderTargets( 1, &iRenderTargetView, iDepthStencilView );
	iEffectModel->GetVariableByName( "InputTexture" )->AsShaderResource()->
					SetResourceArray( iShaderResourceView, 0, 4 );
	iEffectModel->GetTechniqueByName( _TechniqueSort )->
									GetPassByIndex( 0 )->Apply( 0 );
	iDevice->Draw( 4, 0 );


	// Xe[gNAĂ
	ID3D10ShaderResourceView*	_nullSRV[MAX_RENDERBUFFER];
	for( int i= 0 ; i< MAX_RENDERBUFFER ; i++ ){
		_nullSRV[i]= NULL;
	}
	iEffectModel->GetVariableByName( "InputTexture" )->AsShaderResource()->
							SetResourceArray( _nullSRV, 0, MAX_RENDERBUFFER );
	iEffectModel->GetTechniqueByName( _TechniqueSort )->
									GetPassByIndex( 0 )->Apply( 0 );



	iSwapChain->Present( 0, 0 );
}




/*!
	@brief
		蔲main
*/
int APIENTRY
WinMain( HINSTANCE hinstance, HINSTANCE hpinstance, LPSTR cmdline, int show )
{
	Sample*	riSample= new Sample();

	for(; riSample->WaitMessage() ;){
		riSample->Render();
	}

	delete riSample;
	return	0;
}

